# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2007 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

from os.path import expanduser, exists
from xml.dom.minidom import parse, getDOMImplementation
from gobject import GObject, TYPE_PYOBJECT, TYPE_NONE, SIGNAL_RUN_LAST, type_register

from Logger import debug, error
from snippets.Snippet import Snippet
from builder.Job import Job
from builder.Profile import Profile
from util import str_bool
from Installation import Installation


class Settings(GObject):
	"""
	Central settings for the plugin based on a XML file
	
	Simple values are managed with set/get. Complex settings like build profiles
	or snippets have their special getters and setter.
	
	This is implemented as a singleton class.
	"""
	
	# FIXME: the settings are never flushed because we have no cleanup function
	# called by gedit, we have to flush it after every write :(
	# Maybe we could do this in a thread
	#
	# FIXME: we may not use non-ascii chars here, see
	# http://diveintopython.org/xml_processing/unicode.html
	# http://lists.fourthought.com/pipermail/4suite/2004-November/006691.html
	
	__gsignals__ = {
		"profile-updated" : (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_PYOBJECT])), 	# profile
		"profile-deleted" : (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_PYOBJECT])), 	# profile name
		"profile-saved" : (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_PYOBJECT]))		# profile name
	}
	
	def __new__(type):
		if not '_instance' in type.__dict__:
			#type._instance = object.__new__(type)
			type._instance = GObject.__new__(type)
		return type._instance
	
	def __init__(self):
		if not '_ready' in dir(self):
			GObject.__init__(self)
			
			debug(self, "__init__")
			
			self.__filename = Installation().settings_filename
			
			try:
				self.__dom = parse(self.__filename)
			except IOError:
				error(self, "__init__", "File not found: %s, creating new settings DOM" % self.__filename)
				self.__dom = getDOMImplementation().createDocument(None, "settings", None)
				
			# read buffers
			self.__setting_buffer = {}
			self.__profile_buffer = {}
			self.__snippet_buffer = {}
			
			self.__file = None
			
			debug(self, "__init__", "done")
			
			self._ready = True
		
	def __get_setting_node(self, key):
		for node in self.__dom.getElementsByTagName("setting"):
			if node.getAttribute("key") == key:
				return node
		raise KeyError
	
	def __flush(self):
		"""
		Flush settings to XML file
		"""
		try:
			self.__file = open(self.__filename, "w+")
			self.__file.write(self.__dom.toxml())
			self.__file.flush()
			
		except UnicodeDecodeError:
			error(self, "__flush", "UnicodeDecodeError: we currently can't store non-ASCII characters in the XML settings...")
	
	def __getitem__(self, key):
		try:
			value = self.__setting_buffer[key]
			#log(self, "__getitem__(%s) : found in buffer" % (key))
			return value
		except KeyError:
			#log(self, "__getitem__(%s) : buffer KeyError" % (key))
			value = self.__get_setting_node(key).getAttribute("value")
			self.__setting_buffer[key] = value
			return value
	
	def __setitem__(self, key, value):
		try:
			node = self.__get_setting_node(key)
			node.setAttribute("value", str(value))
		except KeyError:
			node = self.__dom.createElement("setting")
			node.setAttribute("key", key)
			node.setAttribute("value", str(value))
			self.__dom.documentElement.appendChild(node)
			
		#log(self, "__setitem__(%s) : buffering" % key)
		self.__setting_buffer[key] = str(value)
		
		# TODO: don't flush if the value has not changed
		self.__flush()
	
	def setdefault(self, key, value):
		"""
		This is __getitem__ but it adds the setting if it's not found and sets
		the default value
		"""
		try:
			return self.__getitem__(key)
		except KeyError:
			self.__setitem__(key, value)
			return value
	
	def __get_profile_node(self, name):
		"""
		Find a profile node by its name
		"""
		for profile_node in self.__dom.getElementsByTagName("profile"):
			if profile_node.getAttribute("name") == name:
				return profile_node
		raise KeyError
	
	def __get_profile_from_node(self, profile_node):
		"""
		Create a Profile object from a <profile> node
		"""
		profile = Profile()
		profile.name = profile_node.getAttribute("name")
		profile.viewer = profile_node.getAttribute("viewer")
		profile.accel = profile_node.getAttribute("accel")
		
		icon = profile_node.getAttribute("icon")
		if icon[0] != "/":
			icon = "%s/%s" % (Installation().path, icon)	# expand relative path
		profile.icon = icon
		
		for job_node in profile_node.getElementsByTagName("job"):
			job = Job()
			job.command = job_node.getAttribute("command")
			job.processor = job_node.getAttribute("processor")
			job.stopper = str_bool(job_node.getAttribute("stopper"))
					
			profile.jobs.append(job)
				
		return profile
	
	def get_profile(self, name):
		"""
		Find a profile by its name
		"""
		return self.__get_profile_from_node(self.__get_profile_node(name))
	
	def list_profiles(self):
		"""
		Return a list of all profiles
		"""
		profiles = []
		
		for profile_node in self.__dom.getElementsByTagName("profile"):
			profile = self.__get_profile_from_node(profile_node)

			self.__profile_buffer[profile.name] = profile
			profiles.append(profile)
		
		return profiles
	
	def update_profile(self, profile):
		"""
		The profile is identified by its name!
		"""
		try:
			profile_node = self.__get_profile_node(profile.name)
			profile_node.setAttribute("viewer", profile.viewer)
			profile_node.setAttribute("accel", profile.accel)
			
			# remove old jobs
			for job_node in profile_node.getElementsByTagName("job"):
				profile_node.removeChild(job_node)
				# TODO: unlink
				
			# append new jobs
			for job in profile.jobs:
				job_node = self.__dom.createElement("job")
				job_node.setAttribute("command", job.command)
				job_node.setAttribute("processor", job.processor)
				job_node.setAttribute("stopper", str(job.stopper))
				profile_node.appendChild(job_node)
			
			# update buffer
			self.__profile_buffer[profile.name] = profile
			self.__flush()
			self.emit("profile-updated", profile)
			
		except KeyError:
			error(self, "update_profile", "Profile not found: %s" % profile.name)
		
	def delete_profile(self, profile):
		try:
			profile_node = self.__get_profile_node(profile.name)
			
			self.__dom.documentElement.removeChild(profile_node)
			# TODO: unlink
			
			# remove from buffer
			try:
				del self.__profile_buffer[profile.name]
			except KeyError:
				pass
			
			self.__flush()
			self.emit("profile-deleted", profile)
			
		except KeyError:
			error(self, "delete_profile", "Profile not found: %s" % profile.name)
	
	def save_profile(self, profile):
		try:
			self.__get_profile_node(profile.name)
			error(self, "save_profile", "a profile named %s already exists" % profile.name)
		except KeyError:
			profile_node = self.__dom.createElement("profile")
			profile_node.setAttribute("name", profile.name)
			profile_node.setAttribute("viewer", profile.viewer)
			profile_node.setAttribute("icon", "pixmaps/compile.png")
			profile_node.setAttribute("accel", profile.accel)
			
			# append jobs
			for job in profile.jobs:
				job_node = self.__dom.createElement("job")
				job_node.setAttribute("command", job.command)
				job_node.setAttribute("processor", job.processor)
				job_node.setAttribute("stopper", str(job.stopper))
				profile_node.appendChild(job_node)
			
			self.__dom.documentElement.appendChild(profile_node)
			
			profile.icon = "%s/pixmaps/compile.png" % Installation().path
			
			self.__profile_buffer[profile.name] = profile
			self.__flush()
			self.emit("profile-saved", profile)
	
	def __get_snippet_node(self, name):
		for snippet_node in self.__dom.getElementsByTagName("snippet"):
			if snippet_node.getAttribute("name") == name:
				return snippet_node
		raise KeyError
	
	def __get_snippet_from_node(self, snippet_node):
		snippet = Snippet()
		
		snippet.name = snippet_node.getAttribute("name")
		snippet.source = snippet_node.getAttribute("source")
		snippet.packages = snippet_node.getAttribute("packages").split(",")
		snippet.contexts = snippet_node.getAttribute("contexts").split(",")
				
		return snippet
	
	def list_snippets(self):
		snippets = []
		
		for snippet_node in self.__dom.getElementsByTagName("snippet"):
			snippet = self.__get_snippet_from_node(snippet_node)
			
			self.__snippet_buffer[snippet.name] = snippet
			snippets.append(snippet)
		
		return snippets
	
	def get_snippet(self, name):
		return self.__get_snippet_from_node(self.__get_snippet_node(name))
	
	def save_snippet(self, snippet):
		try:
			self.__get_snippet_node(snippet.name)
			error(self, "save_snippet", "a snippet named %s already exists" % snippet.name)
		except KeyError:
			snippet_node = self.__dom.createElement("snippet")
			snippet_node.setAttribute("name", snippet.name)
			snippet_node.setAttribute("source", snippet.source)
			snippet_node.setAttribute("packages", ",".join(snippet.packages))
			snippet_node.setAttribute("contexts", ",".join(snippet.contexts))
			
			self.__dom.documentElement.appendChild(snippet_node)
			
			self.__snippet_buffer[snippet.name] = snippet
			self.__flush()
	
	def update_snippet(self, snippet):
		try:
			snippet_node = self.__get_snippet_node(snippet.name)
			snippet_node.setAttribute("source", snippet.source)
			snippet_node.setAttribute("packages", ",".join(snippet.packages))
			snippet_node.setAttribute("contexts", ",".join(snippet.contexts))
			
			# update buffer
			self.__snippet_buffer[snippet.name] = snippet
			self.__flush()
			
		except KeyError:
			error(self, "update_snippet", "Snippet not found: %s " % snippet.name)
	
	def delete_snippet(self, snippet):
		try:
			snippet_node = self.__get_snippet_node(snippet.name)
			
			self.__dom.documentElement.removeChild(snippet_node)
			# TODO: unlink
			
			# remove from buffer
			try:
				del self.__snippet_buffer[snippet.name]
			except KeyError:
				pass
			
			self.__flush()
			
		except KeyError:
			error(self, "delete_snippet", "Snippet not found: %s" % snippet.name)

type_register(Settings)


#class GConfMigrator(object):
#	def __init__(self, settings):
#		self.__settings = settings
#		self.__gconf_root = "/apps/gedit-latex-plugin"
#	
#	def run(self):
#		from gconf import client_get_default, CLIENT_PRELOAD_NONE
#		
#		self.__client = client_get_default()
#		self.__client.add_dir(self.__gconf_root, CLIENT_PRELOAD_NONE)
#		
#		self.__migrate_profiles()
#		self.__migrate_snippets()
#		
#	def __migrate_profiles(self):
#		for path in self.__client.all_dirs("%s/custom-build-profiles" % self.__gconf_root):
#			key = path.split("/")[-1]
#			self.__settings.save_profile(self.__get_profile(key))
#	
#	def __get_profile(self, key):
#		from gconf import VALUE_STRING
#		
#		commands = self.__client.get_list("%s/custom-build-profiles/%s/commands" % (self.__gconf_root, key), VALUE_STRING)
#		
#		profile = Profile()
#		
#		profile.name = self.__client.get_string("%s/custom-build-profiles/%s/name" % (self.__gconf_root, key))
#		
#		viewer = self.__client.get_string("%s/custom-build-profiles/%s/view-command" % (self.__gconf_root, key))
#		if viewer:
#			profile.viewer = viewer
#		
#		accel = self.__client.get_string("%s/custom-build-profiles/%s/accel" % (self.__gconf_root, key))
#		if accel:
#			profile.accel = accel
#			
#		for command in self.__client.get_list("%s/custom-build-profiles/%s/commands" % (self.__gconf_root, key), VALUE_STRING):
#			job = Job()
#			job.command = command
#			job.processor = "default"
#			job.stopper = True
#			
#			profile.jobs.append(job)
#		
#		debug(self, "Profile : " + profile.name)
#		
#		return profile
#	
#	def __migrate_snippets(self):
#		for path in self.__client.all_dirs("%s/user-defined-macros" % self.__gconf_root):
#			key = path.split("/")[-1]
#			self.__settings.save_snippet(self.__get_snippet(key))
#	
#	def __get_snippet(self, key):
#		from gconf import VALUE_STRING
#		
#		snippet = Snippet()
#		
#		snippet.name = self.__client.get_string("%s/user-defined-macros/%s/name" % (self.__gconf_root, key))
#		snippet.source = self.__client.get_string("%s/user-defined-macros/%s/definition" % (self.__gconf_root, key))
#		snippet.packages = self.__client.get_list("%s/user-defined-macros/%s/packages" % (self.__gconf_root, key), VALUE_STRING)
#		c = self.__client.get_int("%s/user-defined-macros/%s/scope" % (self.__gconf_root, key))
#		snippet.context = {0 : "text", 1 : "text", 2 : "math", 3 : "text,math"}[c]
#		
#		debug(self, "Snippet : " + snippet.name)
#		
#		return snippet
#	