# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2007 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

from gtk import Window, TEXT_WINDOW_WIDGET, WINDOW_POPUP, ListStore, TreeView, \
		CellRendererText, ScrolledWindow, POLICY_AUTOMATIC, Frame, SHADOW_OUT
from gtk.gdk import screen_width, screen_height

from proposals import ChoiceProposal
from DocPopup import DocPopup

POPUP_WIDTH = 280
POPUP_HEIGHT = 190
SPACE = 0

class Popup(Window):
	"""
	A popup showing a list of proposals for autocompletion.
	"""
	def __init__(self, window):
		Window.__init__(self, WINDOW_POPUP)
		
		self.__window = window
		
		self.model = ListStore(str, object)		# markup, Proposal instance
		
		self.view = TreeView(self.model)
		self.view.insert_column_with_attributes(-1, "", CellRendererText(), markup=0)
		self.view.set_enable_search(False)
		self.view.set_headers_visible(False)
		
		self.scroll = ScrolledWindow()
		self.scroll.set_policy(POLICY_AUTOMATIC, POLICY_AUTOMATIC)
		self.scroll.add(self.view)
		self.scroll.set_size_request(POPUP_WIDTH, POPUP_HEIGHT)
		
		frame = Frame()
		frame.set_shadow_type(SHADOW_OUT)
		frame.add(self.scroll)
		
		self.add(frame)
		
		self.__doc_popup = DocPopup()
		
		self.__active = False
	
	def set_proposals(self, proposals):
		"""
		Loads proposals into the popup
		"""
		# sort
		proposals.sort(lambda x,y: cmp(str(x), str(y)))
		
		# load
		self.model.clear()
		
		for p in proposals:
			self.model.append([p.label, p])
			
		self.view.set_cursor((0))
	
	@property
	def selected_proposal(self):	
		"""
		Returns the currently selected proposal
		"""
		model, it = self.view.get_selection().get_selected()
		return model.get_value(it, 1)
	
	def navigate(self, key):
		"""
		Moves the selection in the view according to key
		"""
		if key == "Up":
			d = -1
		elif key == "Down":
			d = 1
		elif key == "Page_Up":
			d = -5
		elif key == "Page_Down":
			d = 5
		else:
			return
		
		path = self.view.get_cursor()[0]
		index = path[0]
		max_index = self.model.iter_n_children(None)
		
		index += d
		
		if index < 0:
			index = 0
		elif index >= max_index:
			index = max_index - 1
			
		self.view.set_cursor(index)
		
		self._update_doc_popup()
	
	def _update_doc_popup(self):
		index = self.view.get_cursor()[0][0]
		proposal = self.model[index][1]
		
		if not type(proposal) is ChoiceProposal:
			return
		
		if not proposal.doc:
			return
		
		self.__doc_popup.set_markup(proposal.doc)
		
		x, y = self.get_position()
		width = self.get_size()[0]
		path, column = self.view.get_cursor()
		rect = self.view.get_cell_area(path, column)
		self.__doc_popup.move(x + width + 5, y + rect.y)
		
		self.__doc_popup.set_active(True)
	
	def set_active(self, active):
		"""
		Shows or hides the popup
		"""
		if active and not self.__active:
			self.show_all()
			self._update_doc_popup()
		elif not active and self.__active:
			self.__doc_popup.set_active(False)
			self.hide()
		
		self.__active = active
	
	def get_active(self):
		return self.__active

	def __get_cursor_pos(self, text_view):
		text_buffer = text_view.get_buffer()
		insert_mark = text_buffer.get_insert()
		start = text_buffer.get_iter_at_mark(insert_mark)
		location = text_view.get_iter_location(start)
		win_x, win_y = text_view.buffer_to_window_coords(TEXT_WINDOW_WIDGET, location.x, location.y)
		
		win = text_view.get_window(TEXT_WINDOW_WIDGET)
		xx, yy = win.get_origin()
		
		x = win_x + xx
		y = win_y + yy + location.height + SPACE
		
		return (x, y)
	
	def move_to_cursor(self, text_view):
		x = 0
		y = 0
	
		sw = screen_width()
		sh = screen_height()
		
		x, y = self.__get_cursor_pos(text_view)
		
		w, h = self.get_size()
		
		if x + w > sw:
			x = sw - w - 4
		
		if y + h > sh:
			layout = text_view.create_pango_layout("a")		# ???
			xtext, ytext = layout.get_pixel_size()
			y = y - ytext - h
		
		self.move(x, y)
		
